from sklearn.model_selection import train_test_split
import numpy as np
from medcat.utils.models import LSTM as MODEL
from sklearn.metrics import classification_report, f1_score, confusion_matrix, precision_score, recall_score
import torch
from torch import nn
import torch.nn.functional as F
import torch.optim as optim


def get_batch(ind, batch_size, x, y, cpos, device):
    # Get the start/end index for this batch
    start = ind * batch_size
    end = (ind+1) * batch_size

    # Get the batch
    x_batch = x[start:end]
    y_batch = y[start:end]
    c_batch = cpos[start:end]

    # Return and move the batches to the right device
    return x_batch.to(device), y_batch.to(device), c_batch.to(device)


def train_network(net, data, lr=0.01, test_size=0.1, max_seq_len=41, pad_id=30000, batch_size=100,
                  nepochs=20, device='cpu', save_dir='./meta_cat/', class_weights=None, ignore_cpos=False,
                  auto_save_model=True):
    # Split data
    y = np.array([x[0] for x in data])
    x = [x[1] for x in data]
    cent = np.array([x[2] for x in data])

    # Pad X and convert to array
    x = np.array([(sample + [pad_id] * max(0, max_seq_len - len(sample)))[0:max_seq_len] for sample in x])

    x_train, x_test, y_train, y_test, c_train, c_test = train_test_split(x, y,
            cent, test_size=test_size)

    x_train = torch.tensor(x_train, dtype=torch.long)
    y_train = torch.tensor(y_train, dtype=torch.long)
    c_train = torch.tensor(c_train, dtype=torch.long)

    x_test = torch.tensor(x_test, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)
    c_test = torch.tensor(c_test, dtype=torch.long)

    device = torch.device(device) # Create a torch device
    if class_weights is not None:
        class_weights = torch.FloatTensor(class_weights).to(device)
        criterion = nn.CrossEntropyLoss(weight=class_weights) # Set the criterion to Cross Entropy Loss
    else:
        criterion = nn.CrossEntropyLoss() # Set the criterion to Cross Entropy Loss
    parameters = filter(lambda p: p.requires_grad, net.parameters())
    optimizer = optim.Adam(parameters, lr=lr)
    net.to(device) # Move the network to device

    batch_size = 40
    # Calculate the number of batches given training size len(x_train)
    num_batches = int(np.ceil(len(x_train) / batch_size))
    best_f1 = 0
    best_p = 0
    best_r = 0
    for epoch in range(nepochs):
        running_loss_train = []
        running_loss_test = []

        # TRAIN PHASE
        net.train()
        train_outs = []
        for i in range(num_batches):
            x_train_batch, y_train_batch, cpos_train_batch = get_batch(ind=i,
                                                                       batch_size=batch_size,
                                                                       x=x_train,
                                                                       y=y_train, cpos=c_train,
                                                                       device=device)
            optimizer.zero_grad()
            outputs = net(x_train_batch, cpos_train_batch, ignore_cpos=ignore_cpos)
            loss = criterion(outputs, y_train_batch)
            loss.backward()
            running_loss_train.append(loss.item())

            parameters = filter(lambda p: p.requires_grad, net.parameters())
            torch.nn.utils.clip_grad_norm_(parameters, 0.25)
            optimizer.step()

            train_outs.append(outputs.detach().cpu().numpy())


        # TEST PHASE

        num_batches_test = int(np.ceil(len(x_test) / batch_size))
        test_outs = []
        for j in range(num_batches_test):
            net.eval()
            x_test_batch, y_test_batch, cpos_test_batch = get_batch(ind=j,
                                                                     batch_size=batch_size,
                                                                     x=x_test,
                                                                     y=y_test,
                                                                     cpos=c_test,
                                                                     device=device)
            outputs_test = net(x_test_batch, cpos_test_batch, ignore_cpos=ignore_cpos)
            test_outs.append(outputs_test.detach().cpu().numpy())
            running_loss_test.append(criterion(outputs_test, y_test_batch).item())

        train_loss = np.average(running_loss_train)
        test_loss = np.average(running_loss_test)

        print("*"*50 + "  Train")
        print(classification_report(y_train, np.argmax(np.concatenate(train_outs, axis=0), axis=1)))
        print("*"*50 + "  Test")
        print(classification_report(y_test, np.argmax(np.concatenate(test_outs, axis=0), axis=1)))
        print("Train Loss: {:5}\nTest Loss:  {:5}\n\n".format(train_loss, test_loss))
        print("\n\n\n")
        f1 = f1_score(y_test, np.argmax(np.concatenate(test_outs, axis=0), axis=1), average='weighted')
        precision = precision_score(y_test, np.argmax(np.concatenate(test_outs, axis=0), axis=1), average='weighted')
        recall = recall_score(y_test, np.argmax(np.concatenate(test_outs, axis=0), axis=1), average='weighted')

        if f1 > best_f1:
            print("=" * 50)
            if auto_save_model:
                path = save_dir + "lstm.dat"
                torch.save(net.state_dict(), path)
                print("Model saved at epoch: {} and f1: {}".format(epoch, f1))

            best_f1 = f1
            best_p = precision
            best_r = recall

            # Print some stats
            print(confusion_matrix(y_test, np.argmax(np.concatenate(test_outs, axis=0), axis=1)))
            print("\n\n")

    return (best_f1, best_p, best_r)



import numpy as np

def prepare_from_json(data, cntx_left, cntx_right, tokenizer, lowercase=True, cntx_in_chars=False, tui_filter=None):
    """ Convert the data from a json format into a CSV-like format for training.

    data:  json file from MedCAT
    cntx_left:  size of the context
    cntx_right:  size of the context
    tokenizer:  instance of the <Tokenizer> class from huggingface

    return:  {'category_name': [('category_value', 'tokens', 'center_token'), ...], ...}
    """
    out_data = {}

    for project in data['projects']:
        for document in project['documents']:
            if lowercase:
                text = str(document['text']).lower()
            else:
                text = str(document['text'])

            if len(text) > 0:
                doc_text = tokenizer.encode(text)

                for ann in document['annotations']:
                    tui = ""
                    if tui_filter:
                        tui = ann['tui']

                    if not tui_filter or tui in tui_filter:
                        if ann.get('validated', True) and (not ann.get('deleted', False) and not ann.get('killed', False)):
                            start = ann['start']
                            end = ann['end']

                            if not cntx_in_chars:
                                # Get the index of the center token
                                ind = 0
                                for ind, pair in enumerate(doc_text.offsets):
                                    if start >= pair[0] and start <= pair[1]:
                                        break

                                _start = max(0, ind - cntx_left)
                                _end = min(len(doc_text.tokens), ind + 1 + cntx_right)
                                tkns = doc_text.tokens[_start:_end]
                                cpos = cntx_left + min(0, ind-cntx_left)
                            else:
                                _start = max(0, start - cntx_left)
                                _end = min(len(text), end + cntx_right)
                                tkns = tokenizer.encode(text[_start:_end]).tokens


                            # Backward compatibility if meta_anns is a list vs dict in the new approach
                            meta_anns = []
                            if 'meta_anns' in ann:
                                meta_anns = ann['meta_anns']

                                if type(meta_anns) == dict:
                                    meta_anns = meta_anns.values()

                            # If the annotation is validated
                            for meta_ann in meta_anns:
                                name = meta_ann['name']
                                value = meta_ann['value']

                                sample = [value, tkns, cpos]

                                if name in out_data:
                                    out_data[name].append(sample)
                                else:
                                    out_data[name] = [sample]

    return out_data


def encode_category_values(data, vals=None):
    data = list(data)
    if vals is None:
        vals = set([x[0] for x in data])
        vals = {name:i for i,name in enumerate(vals)}

    # Map values to numbers
    for i in range(len(data)):
        data[i][0] = vals[data[i][0]]

    return data, vals


def tkns_to_ids(data, tokenizer):
    data = list(data)

    for i in range(len(data)):
        data[i][1] = [tokenizer.token_to_id(tok) for tok in data[i][1]]

    return data


def make_mc_train_test(data, cdb, seed=17, test_size=0.2):
    """ This is a disaster
    """
    cnts = {}
    total_anns = 0
    # Count all CUIs
    for project in data['projects']:
        cui_filter = None
        tui_filter = None

        if 'cuis' in project and len(project['cuis'].strip()) > 0:
            cui_filter = [x.strip() for x in project['cuis'].split(",")]
        if 'tuis' in project and len(project['tuis'].strip()) > 0:
            tui_filter = [x.strip().upper() for x in project['tuis'].split(",")]

        for document in project['documents']:
            for ann in document['annotations']:

                if (cui_filter is None and tui_filter is None) or (cui_filter is not None and ann['cui'] in cui_filter) or \
                   (tui_filter is not None and cdb.cui2tui.get(ann['cui'], 'unk') in tui_filter):
                    if ann['cui'] in cnts:
                        cnts[ann['cui']] += 1
                    else:
                        cnts[ann['cui']] = 1

                    total_anns += 1


    test_cnts = {}
    test_anns = 0
    test_prob = 0.90

    test_set = {'projects': []}
    train_set = {'projects': []}

    for i_project in np.random.permutation(np.arange(0, len(data['projects']))):
        project = data['projects'][i_project]
        cui_filter = None
        tui_filter = None

        test_project = {}
        train_project = {}
        for k, v in project.items():
            if k == 'documents':
                test_project['documents'] = []
                train_project['documents'] = []
            else:
                test_project[k] = v
                train_project[k] = v

        if 'cuis' in project and len(project['cuis'].strip()) > 0:
            cui_filter = [x.strip() for x in project['cuis'].split(",")]
        if 'tuis' in project and len(project['tuis'].strip()) > 0:
            tui_filter = [x.strip().upper() for x in project['tuis'].split(",")]


        for i_document in np.random.permutation(np.arange(0, len(project['documents']))):
            # Do we have enough documents in the test set
            if test_anns / total_anns >= test_size:
                test_prob = 0

            document = project['documents'][i_document]

            # Coutn CUIs for this document
            _cnts = {}
            for ann in document['annotations']:
                if (cui_filter is None and tui_filter is None) or (cui_filter is not None and ann['cui'] in cui_filter) or \
                   (tui_filter is not None and cdb.cui2tui.get(ann['cui'], 'unk') in tui_filter):
                    if ann['cui'] in _cnts:
                        _cnts[ann['cui']] += 1
                    else:
                        _cnts[ann['cui']] = 1


            # Did we get more than 30% of concepts for any CUI with >=10 cnt
            is_test = True
            for cui, v in _cnts.items():
                if (v + test_cnts.get(cui, 0)) / cnts[cui] > 0.3:
                    if cnts[cui] >= 10:
                        # We only care for concepts if count >= 10, else they will be ignored
                        #during the test phase (for all metrics and similar)
                        is_test = False
                        break

            # Add to test set
            if is_test and np.random.rand() < test_prob:
                test_project['documents'].append(document)
                for ann in document['annotations']:
                    if (cui_filter is None and tui_filter is None) or (cui_filter is not None and ann['cui'] in cui_filter) or \
                       (tui_filter is not None and cdb.cui2tui.get(ann['cui'], 'unk') in tui_filter):
                        test_anns += 1
                        if ann['cui'] in test_cnts:
                            test_cnts[ann['cui']] += 1
                        else:
                            test_cnts[ann['cui']] = 1
            else:
                train_project['documents'].append(document)

        test_set['projects'].append(test_project)
        train_set['projects'].append(train_project)

    return train_set, test_set, test_anns, total_anns



class Config(object):
    def __init__(self, **kwargs):
        # Is the pipeline in the debug mode or not, debug mainly means it will
        #log a lot of stuff
        self.debug = kwargs.pop("debug", False)
        # What punctuation to NOT ignore during annotation, by default all punct is ignored
        self.keep_punct = kwargs.pop("keep_punct", ['.', ':'])
        # Do we want to have nested entities in the JSON output
        self.nested_entities = kwargs.pop("nested_entities", False)
        # When calculating the context embedding to we want normalized tokens (tkn._.norm) or just lowercased (tkn.text.lower())
        self.norm_emb = kwargs.pop("norm_emb", False)
        # If True, concepts that appear frequently will be prefered during disambiguation
        self.prefer_frequent = kwargs.pop("prefer_frequent", False)
        # If True, concepts that have an ICD10 code will be prefer during disambiguation
        self.prefer_icd10 = kwargs.pop("prefer_icd10", False)
        # Window size when calculating the context embedding (N - from left, N from right)
        self.cntx_span = kwargs.pop("cntx_span", 9)
        # Same as CNTX_SPAN, but usually 1/3 of the size
        self.cntx_span_short = kwargs.pop("cntx_span_short", 3)
        # If a concept appeared below this number it will always be trained in the unsuperwised mode, above this number
        #the concept will be trained only sometimes with decreasing train probability depending on the number of occurences.
        self.min_cui_count = kwargs.pop("min_cui_count", 30000)
        # Minimum number of trainign examples a concept must have before it can be considered 
        #for disambiguation
        self.disambiguation_limit = kwargs.pop("disambiguation_limit", 1)
        # By default medcat will not calculate the accuracy for concepts that are not ambiguous (acc=1 for them), but you can enable
        #that by setting the below to True.
        self.acc_always = kwargs.pop("acc_always", False)
        # Very similar to acc_always, but this setting will disambiguate all concepts no matter are there really
        #ambiguous or not
        self.disamb_everything = kwargs.pop("disamb_everything", False)
        # When detecting entity what is the maximum number of tokens that can be skipped/ignored while still considering 
        #the span to be one entity. Used for multi-word entites that possible can have additional words in the text that we can ignore. 
        self.max_skip_tkn = kwargs.pop("max_skip_tkn", 2)
        # Do we ignore stopwords or not
        self.skip_stopwords = kwargs.pop("skip_stopwords", False)
        # When calculating the context embedding do we want to use weighted average or simply average
        self.weighted_average = kwargs.pop("weighted_average", True)
        # Minimu accuracy for concept detection, concepts below will not be shown/returned
        self.min_acc = kwargs.pop("min_acc", 0.2)
        # Mimum length of the concept span that can be detected
        self.min_concept_length = kwargs.pop("min_concept_length", 1)
        # Probability of adding a negative context each time we add a positive one
        self.neg_prob = kwargs.pop("neg_prob", 0.5)
        # Style of the labels, currently supported long/ent/none
        self.lbl_style = kwargs.pop("lbl_style", 'ent').lower()



import setuptools
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info

with open("./README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="medcat",
    version="0.3.3.7",
    author="w-is-h",
    author_email="w.kraljevic@gmail.com",
    description="Concept annotation tool for Electronic Health Records",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CogStack/MedCAT",
    packages=['medcat', 'medcat.utils', 'medcat.preprocessing'],
    install_requires=[
        'numpy~=1.18',
        'pandas~=1.0',
        'gensim~=3.7',
        'spacy==2.2.4',
        'scipy~=1.4',
        'tokenizers~=0.6.0',
        'torch~=1.4.0',
        'torchvision~=0.5.0',
        'Flask~=1.1',
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
